import warnings
from typing import List, Tuple, Self

import numpy as np

try:
    import OpenGL.GL
    import OpenGL.GLU
    import OpenGL.GLUT
except ImportError:
    warnings.warn('OpenGL module not found, OpenGL backend will not be available.')

from .base_render_backend import RenderBackend, Surface, Font
from .events import Event, EventType


class OpenglFont(Font):
    def __init__(self, font, font_name: str, font_size: int):
        super().__init__(font_name, font_size)
        self.font = font

    def render(self, text: str, color: np.ndarray, antialias: bool, background: np.ndarray = None) -> Surface:
        pass


class OpenglSurface(Surface):
    def __init__(self, surface):
        super().__init__()
        pass

    def get_size(self) -> Tuple[int, int]:
        pass

    def fill(self, color: np.ndarray):
        pass

    def line(self, color: np.ndarray, start: np.ndarray, end: np.ndarray):
        pass

    def circle(self, color: np.ndarray, pos: Tuple[int, int] | np.ndarray, radius: int):
        pass

    def blit(self, surface: Self, pos: Tuple[int, int]):
        pass


class OpenglBackend(RenderBackend):
    def init(self):
        OpenGL.GLUT.glutInit()
        OpenGL.GLUT.glutInitDisplayMode(OpenGL.GLUT.GLUT_RGBA | OpenGL.GLUT.GLUT_DOUBLE)

    def quit(self):
        OpenGL.GLUT.glutLeaveMainLoop()

    def get_font(self, font_size: int, font_name: str = '') -> OpenglFont:
        if font_name == '':
            font = OpenGL.GLUT.GLUT_BITMAP_HELVETICA_12 if font_size <= 12 else OpenGL.GLUT.GLUT_BITMAP_HELVETICA_18
        else:
            font = OpenGL.GLUT.GLUT_BITMAP_HELVETICA_12  # Default to Helvetica if custom font not found
        return OpenglFont(font, font_name, font_size)

    def set_key_repeat(self, delay: int, interval: int):
        # Not directly supported in OpenGL, would need custom implementation
        warnings.warn('set_key_repeat is not supported in OpenGL, ignoring call to set_key_repeat()')

    def create_window(self, title: str, size: Tuple[int, int] = (0, 0)) -> OpenglSurface:
        OpenGL.GLUT.glutInitWindowSize(*size)
        window = OpenGL.GLUT.glutCreateWindow(title)
        OpenGL.GL.glClearColor(0.0, 0.0, 0.0, 0.0)
        OpenGL.GL.glMatrixMode(OpenGL.GL.GL_PROJECTION)
        OpenGL.GL.glLoadIdentity()
        OpenGL.GLU.gluOrtho2D(0, size[0], size[1], 0)
        return OpenglSurface(window)

    def create_surface(self, size: Tuple[int, int], enable_alpha: bool = True) -> Surface:
        # Create an OpenGL texture/framebuffer object
        texture = OpenGL.GL.glGenTextures(1)
        OpenGL.GL.glBindTexture(OpenGL.GL.GL_TEXTURE_2D, texture)
        if enable_alpha:
            OpenGL.GL.glTexImage2D(OpenGL.GL.GL_TEXTURE_2D, 0, OpenGL.GL.GL_RGBA,
                                   size[0], size[1], 0, OpenGL.GL.GL_RGBA,
                                   OpenGL.GL.GL_UNSIGNED_BYTE, None)
        else:
            OpenGL.GL.glTexImage2D(OpenGL.GL.GL_TEXTURE_2D, 0, OpenGL.GL.GL_RGB,
                                   size[0], size[1], 0, OpenGL.GL.GL_RGB,
                                   OpenGL.GL.GL_UNSIGNED_BYTE, None)
        return OpenglSurface(texture)

    def swap_buffers(self):
        OpenGL.GLUT.glutSwapBuffers()

    def get_events(self) -> List[Event]:
        events = []
        while OpenGL.GLUT.glutMainLoopEvent():
            # Process GLUT events and convert them to our Event format
            event = Event(EventType.QUIT)  # Placeholder
            events.append(event)
        return events



from typing import List, Tuple, Self

import numpy as np
import pygame as pg

from .base_render_backend import RenderBackend, Surface, Font
from .events import Event, EventType


class PygameFont(Font):
    def __init__(self, font: pg.font.Font, font_name: str, font_size: int):
        super().__init__(font_name, font_size)
        self.font = font

    def render(self, text: str, color: np.ndarray, antialias: bool, background: np.ndarray = None) -> Surface:
        return PygameSurface(self.font.render(text, antialias, color, background))


class PygameSurface(Surface):
    def __init__(self, surface: pg.Surface):
        super().__init__()
        self.surface = surface

    def get_size(self) -> Tuple[int, int]:
        return self.surface.get_size()

    def fill(self, color: np.ndarray):
        self.surface.fill(color)

    def line(self, color: np.ndarray, start: np.ndarray, end: np.ndarray):
        pg.draw.line(self.surface, color, start, end)

    def circle(self, color: np.ndarray, pos: Tuple[int, int] | np.ndarray, radius: int):
        pg.draw.circle(self.surface, color, pos, radius)

    def blit(self, surface: Self, pos: Tuple[int, int]):
        self.surface.blit(surface.surface, pos)


class PygameBackend(RenderBackend):
    def init(self):
        pg.init()

    def quit(self):
        pg.quit()

    def get_font(self, font_size: int, font_name: str = '') -> PygameFont:
        font_name = font_name or pg.font.get_default_font()
        return PygameFont(pg.font.Font(font_name, font_size), font_name, font_size)

    def set_key_repeat(self, delay: int, interval: int):
        pg.key.set_repeat(delay, interval)

    def create_window(self, title: str, size: Tuple[int, int] = (0, 0)) -> PygameSurface:
        mode = pg.FULLSCREEN if size == (0, 0) else pg.RESIZABLE
        return PygameSurface(pg.display.set_mode(size, mode))

    def create_surface(self, size: Tuple[int, int], enable_alpha: bool = True) -> Surface:
        flags = 0
        if enable_alpha:
            flags = flags | pg.SRCALPHA
        return PygameSurface(pg.Surface(size, flags))

    def swap_buffers(self):
        pg.display.flip()

    def get_events(self) -> List[Event]:
        events = [_map_event(e) for e in pg.event.get()]
        return [e for e in events if e is not None]


def _map_event(event: pg.event.Event) -> Event | None:
    if event.type == pg.QUIT:
        return Event(EventType.QUIT)
    elif event.type == pg.KEYDOWN:
        return Event(EventType.KEYDOWN)
    elif event.type == pg.KEYUP:
        return Event(EventType.KEYUP)
    elif event.type == pg.MOUSEBUTTONDOWN:
        return Event(
            EventType.MOUSEBUTTONDOWN,
            button=event.button,
            mouse_pos=np.array(event.pos, dtype=np.int32)
        )
    elif event.type == pg.MOUSEBUTTONUP:
        return Event(EventType.MOUSEBUTTONUP)
    elif event.type == pg.MOUSEMOTION:
        return Event(
            EventType.MOUSEMOTION,
            mouse_pos=np.array(event.pos, dtype=np.int32),
            mouse_rel=np.array(event.rel, dtype=np.int32)
        )
    elif event.type == pg.WINDOWENTER:
        return Event(EventType.WINDOWENTER)
    elif event.type == pg.WINDOWFOCUSGAINED:
        return Event(EventType.WINDOWFOCUSGAINED)
    elif event.type == pg.WINDOWRESIZED:
        return Event(EventType.WINDOWRESIZED)
    elif event.type == pg.MOUSEWHEEL:
        return Event(EventType.MOUSEWHEEL, scroll=event.y)
    else:
        return None



from typing import Iterable, Tuple, Dict

import numpy as np

from viztools.coordinate_system import CoordinateSystem
from viztools.drawable import Drawable, _normalize_color
from viztools.render_backend.base_render_backend import RenderBackend, Surface
from viztools.render_backend.events import Event, EventType


class Points(Drawable):
    def __init__(
            self, points: np.ndarray, size: int | float | Iterable[int | float] = 3,
            color: np.ndarray | None = None
    ):
        """
        Drawable to display a set of points.
        :param points: A list of points with the shape [N, 2] where N is the number of points.
        :param size: The radius of the points. If set to an integer, this is the radius on the screen in pixels. If set
                     to a float, this is the radius on the screen in units of the coordinate system. If set to a list,
                     it contains the sizes for each point.
        :param color: The color of the points.
        """
        # points
        if not isinstance(points, np.ndarray):
            raise TypeError(f'points must be a numpy array, not {type(points)}.')
        if points.ndim != 2 or points.shape[1] != 2:
            raise ValueError(f'points must be numpy array with shape (N, 2), not {points.shape}.')
        n_points = len(points)
        self._points = points

        # size
        if isinstance(size, (int, float)):
            is_relative_size = isinstance(size, float)
            size = np.repeat(np.array([[size, float(is_relative_size)]], dtype=np.float32), n_points, axis=0)
        elif isinstance(size, np.ndarray):
            if size.shape != (n_points,):
                raise ValueError(f'size must be a numpy array with shape ({n_points},), not {size.shape}.')
            is_relative_size = np.full(n_points, np.issubdtype(size.dtype, np.floating), dtype=np.float32)
            size = np.stack([size.astype(np.float32), is_relative_size], axis=1)
        elif isinstance(size, list):
            if len(size) != n_points:
                raise ValueError(f'size must be a list of length {n_points}, not {len(size)}.')
            size = [[s, isinstance(s, float)] for s in size]
            size = np.array(size, dtype=np.float32)
        else:
            raise TypeError(f'size must be an integer, float or iterable, not {type(size)}.')
        self._size = size

        # colors
        if color is None:
            color = np.array([77, 178, 11])
        if isinstance(color, np.ndarray):
            if color.shape == (3,):
                color = np.array([*color, 255], dtype=np.float32)
            if color.shape == (4,):
                color = np.repeat(color.reshape(1, -1), n_points, axis=0).astype(np.float32)
        else:
            color = np.array(color, dtype=np.float32)
        if color.shape != (n_points, 4):
            raise ValueError(f'colors must be a numpy array with shape ({n_points}, 4), not {color.shape}.')
        self._colors = color

        self._surface_parameters = {}
        for surf_params in self._get_surf_params():
            self._surface_parameters[surf_params.tobytes()] = surf_params

    def __len__(self):
        return len(self._points)

    def _get_surf_params(self) -> np.ndarray:
        return np.concatenate([self._size, self._colors], axis=1)

    def _get_surf_param(self, index: int) -> np.ndarray:
        return np.concatenate([self._size[index, :], self._colors[index, :]], axis=0)

    def set_color(self, color: np.ndarray | Tuple[int, int, int, int], index: int):
        self._colors[index, :] = _normalize_color(color)
        self._update_surf_params(index)

    def _update_surf_params(self, index: int):
        surf_params = self._get_surf_param(index)
        self._surface_parameters[surf_params.tobytes()] = surf_params

    def set_size(self, size: int | float, index: int):
        self._size[index, 0] = size
        self._size[index, 1] = isinstance(size, float)
        self._update_surf_params(index)

    def _create_point_surfaces(self, render_backend: RenderBackend, zoom_factor: float) -> Dict[bytes, Surface]:
        surfaces = {}
        for k, surf_params in self._surface_parameters.items():
            draw_size = _get_draw_size(surf_params[0], zoom_factor, bool(surf_params[1]))
            color = surf_params[2:]

            # old version with per pixel alpha
            point_surface = render_backend.create_surface((draw_size * 2, draw_size * 2), enable_alpha=True)
            point_surface.circle(color, (draw_size, draw_size), draw_size)

            surfaces[k] = point_surface
        return surfaces

    def _get_draw_sizes(self, zoom_factor: float) -> np.ndarray:
        """
        Computes the draw sizes for the given sizes and coordinate system.
        :param zoom_factor: A float defining the scale factor for relative sizes.
        size[i] must be multiplied with zoom_factor.
        :return: numpy array of integers of shape [N,] where N is the number of sizes.
        """
        draw_sizes = self._size[:, 0].copy()
        is_relative_size = self._size[:, 1] > 0.5
        draw_sizes[is_relative_size] *= zoom_factor
        return np.maximum(draw_sizes.astype(int), 1)

    def draw(
            self, screen: Surface, coordinate_system: CoordinateSystem, screen_size: np.ndarray,
            render_backend: RenderBackend
    ):
        draw_sizes = self._get_draw_sizes(coordinate_system.zoom_factor)

        # filter out invalid positions
        screen_points = coordinate_system.space_to_screen(self._points.T).T
        valid_positions = _get_valid_positions(screen_points, draw_sizes, screen_size)
        screen_points = screen_points[valid_positions]
        valid_colors = self._colors[valid_positions]
        valid_sizes = draw_sizes[valid_positions]
        screen_points -= valid_sizes.reshape(-1, 1)

        # create blit surfaces
        surfaces = self._create_point_surfaces(render_backend, coordinate_system.zoom_factor)

        # draw
        for pos, size, color, surf_params in zip(
                screen_points, valid_sizes, valid_colors, self._get_surf_params()[valid_positions]
        ):
            surface = surfaces[surf_params.tobytes()]
            screen.blit(surface, pos)

    def clicked_points(self, event: Event, coordinate_system: CoordinateSystem) -> np.ndarray:
        """
        Returns the indices of the points clicked by the mouse. Returns an empty array if no point was clicked.

        :param event: The event to check.
        :param coordinate_system: The coordinate system to use.
        """
        if event.type == EventType.MOUSEBUTTONDOWN and event.button == 1:
            draw_size = self._get_draw_sizes(coordinate_system.zoom_factor)

            screen_pos = event.mouse_pos.reshape(1, 2)
            screen_points = coordinate_system.space_to_screen(self._points.T).T
            distances = np.linalg.norm(screen_points - screen_pos, axis=1)
            return np.nonzero(distances < draw_size)[0]
        return np.array([])


def _get_draw_size(
        size: float, zoom_factor: float, is_relative_size: bool
) -> int:
    if is_relative_size:
        size = max(int(size * zoom_factor), 1)
    return int(size)


def _get_valid_positions(screen_points: np.ndarray, draw_sizes: np.ndarray, screen_size: np.ndarray) -> np.ndarray:
    return np.where(np.logical_and(
        np.logical_and(screen_points[:, 0] > -draw_sizes, (screen_points[:, 0] < screen_size[0] + draw_sizes)),
        np.logical_and(screen_points[:, 1] > -draw_sizes, (screen_points[:, 1] < screen_size[1] + draw_sizes))
    ))[0]



from abc import ABC, abstractmethod
from typing import Tuple

import numpy as np

from viztools.coordinate_system import CoordinateSystem
from viztools.render_backend.base_render_backend import Surface, RenderBackend

Color = np.ndarray | Tuple[int, int, int, int] | Tuple[int, int, int]


class Drawable(ABC):
    @abstractmethod
    def draw(
            self, screen: Surface, coordinate_system: CoordinateSystem, screen_size: np.ndarray,
            render_backend: RenderBackend
    ):
        pass


def _normalize_color(color: Color) -> np.ndarray:
    if len(color) == 3:
        return np.array([*color, 255], dtype=np.float32)
    if len(color) != 4:
        raise ValueError(f'color must be of length 3 or 4, not {len(color)}.')
    return np.array(color, dtype=np.float32)





from enum import StrEnum

import numpy as np


class EventType(StrEnum):
    QUIT = 'QUIT'
    KEYDOWN = 'KEYDOWN'
    KEYUP = 'KEYUP'
    MOUSEBUTTONDOWN = 'MOUSEBUTTONDOWN'
    MOUSEBUTTONUP = 'MOUSEBUTTONUP'
    MOUSEMOTION = 'MOUSEMOTION'
    WINDOWENTER = 'WINDOWENTER'
    WINDOWFOCUSGAINED = 'WINDOWFOCUSGAINED'
    WINDOWRESIZED = 'WINDOWRESIZED'
    MOUSEWHEEL = 'MOUSEWHEEL'


class Event:
    def __init__(
            self, event_type: EventType, mouse_pos: np.ndarray | None = None, mouse_rel: np.ndarray | None = None,
            scroll: int | None = None, button: int | None = None,
    ):
        self.type = event_type
        self.mouse_pos = mouse_pos
        self.mouse_rel = mouse_rel
        self.scroll = scroll
        self.button = button



from abc import ABC, abstractmethod
from typing import Tuple, List, Self

import numpy as np

from .events import Event


class Surface(ABC):
    @abstractmethod
    def get_size(self) -> Tuple[int, int]:
        pass

    @abstractmethod
    def fill(self, color: np.ndarray):
        pass

    @abstractmethod
    def line(self, color: np.ndarray, start: Tuple[int, int] | np.ndarray, end: Tuple[int, int] | np.ndarray):
        pass

    @abstractmethod
    def circle(self, color: np.ndarray, pos: Tuple[int, int] | np.ndarray, radius: int | float):
        pass

    @abstractmethod
    def blit(self, surface: Self, pos: Tuple[int, int] | np.ndarray):
        pass


class Font(ABC):
    def __init__(self, font_name: str, font_size: int):
        self.font_name = font_name
        self.font_size = font_size

    @abstractmethod
    def render(self, text: str, color: np.ndarray, antialias: bool, background: np.ndarray = None) -> Surface:
        pass


class RenderBackend(ABC):
    def __init__(self):
        pass

    @abstractmethod
    def init(self):
        pass

    @abstractmethod
    def quit(self):
        pass

    @abstractmethod
    def get_font(self, font_size: int, font_name: str) -> Font:
        pass

    @abstractmethod
    def set_key_repeat(self, delay: int, interval: int):
        pass

    @abstractmethod
    def create_window(self, title: str, size: Tuple[int, int] = (0, 0)) -> Surface:
        """
        Create a window with the given title and size. If size is (0, 0), the window will be full screen.
        :param title: The title of the window.
        :param size: The size of the window as (width, height).
        """
        pass

    @abstractmethod
    def create_surface(self, size: Tuple[int, int], enable_alpha: bool = True) -> Surface:
        pass

    @abstractmethod
    def swap_buffers(self):
        pass

    @abstractmethod
    def get_events(self) -> List[Event]:
        pass
